-----------------------------------------------------------------------------
-- RapidCast - casting bar lag adjustment addon
-----------------------------------------------------------------------------

-- TODO: Add support for eCastingBar, etc. etc.

if eCastingBar ~= nil then
	function RapidCast_UpdateCastingBarFrame(startTime, duration, now)
		local endTime = startTime + duration;
		if (now > endTime) then
			now = endTime;
		end

		eCastingBar.startTime = startTime;
		eCastingBar.maxValue = endTime;
		eCastingBarStatusBar:SetMinMaxValues(startTime, endTime)
	end

	function RapidCast_SpellcastStart()
	end

	function RapidCast_SpellcastChannelStart()
	end

	function RapidCast_SpellcastDelayed()
	end

elseif aUF and aUF.db.profile.units.player.CastBar then
	-- The difficulty with aUF is that the events can be registered at any time after
	-- startup, or even in between event triggers if the user selects various interface options
	--

	local function RapidCast_GetPlayerAUFUnit()
		
		for obj,used in pairs(aUF.unitid.player) do
			if used then 
				return obj 
			end
		end
	
		return nil
	end

	RapidCast_UpdateCastingBarFrame = function(startTime, duration, now)
		if (startTime > now) then
			startTime = now;
		end
		local endTime = startTime + duration;
		if (now > endTime) then
			now = endTime;
		end

		local self = RapidCast_GetPlayerAUFUnit()

		self.spellCastData.startTime = startTime;
		self.spellCastData.endTime = endTime;

		self.CastBar:SetMinMaxValues(startTime, endTime);
		if (self.spellCastData.casting) then
			self.CastBar:SetValue(now);
		elseif (self.spellCastData.channeling) then
			self.CastBar:SetValue(endTime - (now-startTime));
		end
	end

	RapidCast_SpellcastStart = function() 
		local self = RapidCast_GetPlayerAUFUnit()
		self:ObjectUnregisterEvent("UNIT_SPELLCAST_START") 
		self:SpellcastStart() 
	end

	RapidCast_SpellcastChannelStart = function() 
		local self = RapidCast_GetPlayerAUFUnit()
		self:ObjectUnregisterEvent("UNIT_SPELLCAST_CHANNEL_START") 
		self:SpellcastChannelStart() 
	end

	RapidCast_SpellcastDelayed = function() 
		local self = RapidCast_GetPlayerAUFUnit()
		self:ObjectUnregisterEvent("UNIT_SPELLCAST_DELAYED") 
		self:SpellcastDelayed() 
		self.spellCastData.delay = self.spellCastData.delay - RapidCast.lag
	end

else
	function RapidCast_UpdateCastingBarFrame(startTime, duration, now)
		local endTime = startTime + duration;
		if (now > endTime) then
			now = endTime;
		end

		CastingBarFrame.startTime = startTime;
		CastingBarFrame.endTime = endTime;
		CastingBarFrame.maxValue = endTime;
		CastingBarFrame:SetMinMaxValues(startTime, endTime);
		CastingBarFrame:SetValue(now);
	end

	function RapidCast_SpellcastStart()
	end

	function RapidCast_SpellcastChannelStart()
	end

	function RapidCast_SpellcastDelayed()
	end

end

-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("RapidCast")

RapidCast = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0")

RapidCast:RegisterDB("RapidCastDB")
RapidCast:RegisterDefaults('profile', 
	{
		castTimePadding = 0.1,
		incorporateGCD = true,
	}
)

local consoleOptions = {
	type='group',
	args = {
		padding = {
			type = 'range',
			name = L["Cast Time Padding"],
			desc = L["Adjusts the time padding used as a safety margin for casting"],
			get = function() return RapidCast.db.profile.castTimePadding end,
			set = function(v) RapidCast.db.profile.castTimePadding = v end,
			min = 0.0,
			max = 1.0,
			step = 0.01,
		},
		incorporateGCD = {
			type = 'toggle',
			name = L["Incorporate Global Cooldown"],
			desc = L["Adjusts the casting bar to a minimum of 1.5 seconds from when it was cast"],
			get = function() return RapidCast.db.profile.incorporateGCD end,
			set = function(v) RapidCast.db.profile.incorporateGCD = v end,
		},
	},
}

RapidCast:RegisterChatCommand({"/rapidcast"}, consoleOptions)

function RapidCast:OnInitialize()
	self.triggerTime = nil
end

function RapidCast:OnEnable()
	self:RegisterEvent("UNIT_SPELLCAST_SENT", "SpellcastStartTrigger")
	self:RegisterEvent("UNIT_SPELLCAST_START", "SpellcastStartResponse")
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", "SpellcastChannelStartResponse")
	self:RegisterEvent("UNIT_SPELLCAST_DELAYED", "SpellcastDelayed")
	self:RegisterEvent("UNIT_SPELLCAST_STOP", "SpellcastStop")
end

function RapidCast:OnDisable()
end

function RapidCast:SpellcastStartTrigger()
	self.triggerTime = GetTime()
	self.lag = 0
end

function RapidCast:SpellcastStop()
	self.lag = 0
end

function RapidCast:UpdateCastingBars(startTime, endTime, now)
	local duration = (endTime - startTime)*0.001

	if self.db.profile.incorporateGCD then
		if self.triggerTime and endTime*0.001 - self.triggerTime - self.lag < 1.5 then
			duration = 1.5
		end
	end

	RapidCast_UpdateCastingBarFrame(startTime*0.001 - self.lag, duration, now)
end

function RapidCast:StartSpellCommon(startTime, endTime)
	local now = GetTime()

	if now > endTime*0.001 then 
		self.triggerTime = nil
		return 
	end
	
	self.lag = now - self.triggerTime - self.db.profile.castTimePadding
	if self.lag < 0 then self.lag = 0 end

	self:UpdateCastingBars(startTime, endTime, now)

	self.triggerTime = nil
end

function RapidCast:SpellcastStartResponse()
	if arg1 ~= "player" then return end

	RapidCast_SpellcastStart()

	if not self.triggerTime then return end

	local _, _, _, _, startTime, endTime = UnitCastingInfo("player");
	self:StartSpellCommon(startTime, endTime)
end

function RapidCast:SpellcastDelayed()
	if arg1 ~= "player" then return end

	RapidCast_SpellcastDelayed()

	local now = GetTime()
	local _, _, _, _, startTime, endTime = UnitCastingInfo("player")

	self:UpdateCastingBars(startTime, endTime, now)
end


function RapidCast:SpellcastChannelStartResponse()
	if arg1 ~= "player" then return end

	RapidCast_SpellcastChannelStart()

	if not self.triggerTime then return end

	local _, _, _, _, startTime, endTime = UnitChannelInfo("player");
	self:StartSpellCommon(startTime, endTime)
end

-----------------------------------------------------------------------------
